/*
 * mtdnwflash - NetWinder flash device
 *
 * Copyright (c) 2000 Rebel.com
 *
 * This code is GPL
 *
 * Note: The kernel access the route tables from flash. We need
 *       to make sure that we always leave the page where it was.
 *       There should be tons-o-spinlocks here for multiproc
 *       systems.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/mtd/compatmac.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>

typedef struct {
	struct pci_dev *pmu;	/* for switching pages */
	struct pci_dev *south;	/* for enabling writes */
	u8 *flash;
} nw_private;

#define FLASH_ADDR		0xfff80000
#define FLASH_PAGE		0x80000

/* This must be global for the cleanup routine */
static struct mtd_info *mtd_info = NULL;

static int nw_goto_page(struct pci_dev *pmu, int page)
{
	u8 reg;

	if (page > 7) {
		printk("BAD PAGE %d\n", page);
		return 1;	/* bad page */
	}

	pci_read_config_byte(pmu, 0x98, &reg);
	reg = (reg & 3) | ((page ^ 0x3c) << 2);
	pci_write_config_byte(pmu, 0x9d, reg);

	return 0;
}

static int nw_current_page(struct pci_dev *pmu)
{
	u8 reg;

	/*
	 * Extract upper 6 address bits (EGPIO[15:10],
	 * of which topmost 4 are have inverted polarity
	 */
	pci_read_config_byte(pmu, 0x98, &reg);
	reg = ((reg >> 2) & 0x3f) ^ 0x3c;
	if (reg > 7) {
		printk ("Bad Page number, using default\n");
		reg = 3;
	}
	return reg;
}

static void dump_status_reg(u8 value)
{
	printk("Status register 0x%02x:\n", value);
	if (value & 0x80)
		printk("- write state machine busy\n");
	if (value & 0x40)
		printk("- block erase suspended\n");
	if (value & 0x20)
		printk("- error in block erase or clear lock-bit\n");
	if (value & 0x10)
		printk("- error in setting lock-bit\n");
	if (value & 0x08)
		printk("- low programming voltage detected, aborting\n");
	if (value & 0x04)
		printk("- programming has been suspended\n");
	if (value & 0x02)
		printk("- block is locked, aborting\n");
	if (value & 0x01)
		printk("- reserved bit\n");
}

/* SAM Maybe we should not poll here */
static int wait_for_ready(volatile u8 * flash, u8 cmd)
{
	int i;

	for (i = 0; i < 1000000; ++i) {
		*flash = cmd;
		if ((*flash & 0x80) == 0x80)
			return 0;
		udelay(2);
	}

	printk("TIMEOUT\n");
	dump_status_reg(*flash);
	return -EBUSY;
}

inline void write_enable(struct pci_dev *south)
{
	u8 reg;

	pci_read_config_byte(south, 0x47, &reg);
	pci_write_config_byte(south, 0x47, reg | 0x40);
}

inline void write_disable(struct pci_dev *south)
{
	u8 reg;

	pci_read_config_byte(south, 0x47, &reg);
	pci_write_config_byte(south, 0x47, reg & ~0x40);
}

/*
 * Read a single bank (512k) from the flash device.
 */
static int ram_read_bank(struct mtd_info *mtd, loff_t from, 
			 size_t len, size_t * retlen, u_char * buf)
{
	nw_private *priv = mtd_info->priv;
	int page, offset, was;
	size_t max;

	if (from + len > mtd->size)
		return -EINVAL;

	page = from / FLASH_PAGE;
	offset = from & (FLASH_PAGE - 1);
	max = FLASH_PAGE - offset;

	was = nw_current_page(priv->pmu);
	nw_goto_page (priv->pmu, page);

	len = len > max ? max : len;
	// printk ("page %d, memcpy (%p, %p, %x)\n", 
	//	page, buf, priv->flash + offset, len);
	memcpy(buf, priv->flash + offset, len);

	nw_goto_page(priv->pmu, was);

	*retlen = len;
	return 0;
}

/*
 * Read generic from flash (spanning multiple banks)
 */
static int ram_read(struct mtd_info *mtd, loff_t from, size_t len,
		    size_t * retlen, u_char * buf)
{
	loff_t bytes_read = 0;

	while (bytes_read < len) {
		int retlen;
		int result;
		result = ram_read_bank(mtd, from + bytes_read,
				       len - bytes_read, &retlen,
				       buf + bytes_read);
		if (result != 0)
			return result;
		bytes_read += retlen;
	}

	*retlen = bytes_read;
	return 0;
}

/*
 * This routine currently only lets you write to one page at a time.
 * The `to' offset must be 64 byte aligned.
 */
static int ram_write(struct mtd_info *mtd, loff_t to, size_t len,
		     size_t * retlen, const u_char * buf)
{
	nw_private *priv = mtd_info->priv;
	int page, offset;
	size_t max, i, n, was;
	volatile u8 *block;

	if (to & 0x1f)
		return -EINVAL;
	if (to + len > mtd->size)
		return -EINVAL;

	page = to / FLASH_PAGE;
	offset = to & (FLASH_PAGE - 1);
	max = FLASH_PAGE - offset;
	block = priv->flash + offset;

	if(len > max) len = max;

	was = nw_current_page(priv->pmu);
	nw_goto_page(priv->pmu, page);

	write_enable(priv->south);

	for (n = len; n; n -= max) {
		if (wait_for_ready(block, 0xe8)) {	/* block write cmd */
			*priv->flash = 0xff;	/* put back into read mode */
			write_disable(priv->south);
			return -EBUSY;
		}

		max = n < 0x20 ? n : 0x20;

		*block = max - 1;	/* count - 1 */

		for (i = 0; i < max; ++i)
			block[i] = buf[i];

		*block = 0xd0;	/* program */

		block += max;
		buf += max;
	}
	
	--block;

	if (wait_for_ready(block, 0x70)) {	/* status */
		write_disable(priv->south);
		return -EBUSY;
	}

	*priv->flash = 0xff;	/* put back into read mode */

	write_disable(priv->south);

	nw_goto_page(priv->pmu, was);

	*retlen = len;

	return 0;
}

/*
 * block erase: address must be on a 128k boundry
 *              size must be 128k
 */
int ram_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	nw_private *priv = mtd_info->priv;
	int page, offset, was;
	volatile u8 *block;
	u_long addr, len;

	addr = instr->addr;
	len  = instr->len;
	
	if(addr + len > mtd->size) {
		printk("ending address exceeds flash size, aborting\n");
		return -EINVAL;
	}
	if(addr & (mtd->erasesize - 1)) {
		printk("erase address not aligned on block boundary\n");
		return -EINVAL;
	}
	if(!len || len % mtd->erasesize) {
		printk ("erase length not a multiple of block size\n");
		return -EINVAL;
	}

	write_enable(priv->south);

	was = nw_current_page(priv->pmu);

	while(len) {
		page   = addr / FLASH_PAGE;
		offset = addr & (FLASH_PAGE - 1);
		block  = priv->flash + offset;

		nw_goto_page(priv->pmu, page);

		// block erase
		*block = 0x20;
		*block = 0xd0;

		if(wait_for_ready(block, 0x70)) {
			write_disable(priv->south);
			printk("timeout in flash erase\n");
			return -EBUSY;
		}

		addr += mtd->erasesize;
		len  -= mtd->erasesize;
	}

	*block = 0xff;

	nw_goto_page(priv->pmu, was);

	write_disable(priv->south);

	if (instr->callback)
		(*(instr->callback)) (instr);

	return 0;
}


/*
 * Mapping routines.  Only read/write 8/16 are needed.
 */

__u8 nwflash_read8(struct map_info * map, unsigned long ofs)
{
	__u8 t = readb(map->map_priv_1 + ofs);
	// printk("nwflash_read8 %lx = %x\n", ofs, t);
	return t;
}

__u16 nwflash_read16(struct map_info * map, unsigned long ofs)
{
	__u16 t = readw(map->map_priv_1 + ofs);
	// printk("nwflash_read16 %lx = %x\n", ofs, t);
	return t;
}

__u32 nwflash_read32(struct map_info * map, unsigned long ofs)
{
	__u32 t = readl(map->map_priv_1 + ofs);
	// printk("nwflash_read32 %lx = %x\n", ofs, t);
	return t;
}

void nwflash_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	// printk("nwflash_copy from from %lx size %x\n", from, len);
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void nwflash_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	// printk("nwflash_write8 %lx = %x\n", adr, d);
	writeb(d, map->map_priv_1 + adr);
}

void nwflash_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	// printk("nwflash_write16 %lx = %x\n", adr, d);
	writew(d, map->map_priv_1 + adr);
}

void nwflash_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	// printk("nwflash_write32 %lx = %x\n", adr, d);
	writel(d, map->map_priv_1 + adr);
}

void nwflash_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	// printk("nwflash_copy_to %lx size %x\n", to, len);
	memcpy_toio(map->map_priv_1 + to, from, len);
}

#ifdef MODULE
extern struct module __this_module;
#endif

struct map_info nwflash_map = {
	"mtdnwflash",
	FLASH_PAGE,		/* safe minimum */
	2,			/* the probe *only* works in word mode */
	nwflash_read8,
	nwflash_read16,
	nwflash_read32,
	nwflash_copy_from,
	nwflash_write8,
	nwflash_write16,
	nwflash_write32,
	nwflash_copy_to,
	0,
	0
};

mod_init_t init_mtdnwflash(void)
{
	nw_private *priv;
	struct pci_dev *south, *pmu;
	u8 *flash;
	int was;
	struct mtd_info *cfi;
	struct mtd_info *(*probe_function)(struct map_info *);
	char probename[] = "cfi_probe";

	/* Setup */
	if (!(south = pci_find_slot(0, PCI_DEVFN(0x7, 0)))) {
		printk("You do not have a Southbridge\n");
		return 1;
	}

	pci_write_config_dword(south, 0x78, 0xe0000000);

	/* setup the PMU */
	if (!(pmu = pci_find_slot(0, PCI_DEVFN(0x11, 0)))) {
		printk("You do not have a PMU\n");
		return 1;
	}

	pci_write_config_word(pmu, 0x9a, 0xfc00);

	if (!(flash = ioremap(FLASH_ADDR, FLASH_PAGE))) {
		printk("ioremap failed!!!\n");
		return 1;
	}

	printk("flash = %p\n", flash);

	nwflash_map.map_priv_1 = (unsigned long)flash;

	/*
	 * Use the cfi probe routines to get the sizes. We cannot use
	 * the returned functions since they are word oriented and we
	 * are byte oriented.
	 */
	write_enable(south);
	was = nw_current_page(pmu);
	nw_goto_page(pmu, 0);
	probe_function = inter_module_get_request(probename, probename);
	if (!probe_function) {
		printk("Unable to locate the probe function\n");
		iounmap(flash);
		return -EIO;
	}
	cfi = (*probe_function)(&nwflash_map);
	inter_module_put(probename);
	nw_goto_page(pmu, was);
	write_disable(south);
	if (cfi == NULL) {
		printk("UNABLE TO PROBE\n");
		iounmap(flash);
		return -EIO;
	}

	/* Allocate some memory */
	mtd_info = kmalloc(sizeof(struct mtd_info), GFP_KERNEL);
	priv = kmalloc(sizeof(nw_private), GFP_KERNEL);
	if (mtd_info == NULL || priv == NULL) {
		if (mtd_info)
			kfree(mtd_info);
		iounmap(flash);
		return -ENOMEM;
	}

	/* Copy the fields from the cfi */
	memset(mtd_info, 0, sizeof(struct mtd_info));
	mtd_info->name = cfi->name;
	mtd_info->type = cfi->type;
	mtd_info->flags = cfi->flags;
	mtd_info->size = cfi->size;
	mtd_info->erasesize = cfi->erasesize;

	/* Note: Do not free cfi->priv, it points to our nwflash_map! */
	kfree(cfi);

#ifdef MODULE
	mtd_info->module = THIS_MODULE;
#endif

	mtd_info->priv = priv;

	priv->pmu = pmu;
	priv->south = south;
	priv->flash = flash;

	mtd_info->read = ram_read;
	mtd_info->write = ram_write;
	mtd_info->erase = ram_erase;

	if (add_mtd_device(mtd_info)) {
		kfree(mtd_info->priv);
		kfree(mtd_info);
		iounmap(flash);
		return -EIO;
	}

	return 0;
}

mod_exit_t cleanup_mtdnwflash(void)
{
	if (mtd_info) {
		nw_private *priv = mtd_info->priv;

		del_mtd_device(mtd_info);

		iounmap(priv->flash);
		kfree(priv);

		kfree(mtd_info);
		mtd_info = NULL;
	}
}

#ifdef MODULE
module_init(init_mtdnwflash);
module_exit(cleanup_mtdnwflash);
#endif

#if 0
/*
 * Diagnostic routines
 */
void dump_mtd(struct mtd_info *mtd)
{
	long flags;

	printk("%s ", mtd->name);
	switch (mtd->type) {
	case MTD_ABSENT:
		printk("MTD_ABSENT:");
		break;
	case MTD_RAM:
		printk("MTD_RAM:");
		break;
	case MTD_ROM:
		printk("MTD_ROM:");
		break;
	case MTD_NORFLASH:
		printk("MTD_NORFLASH:");
		break;
	case MTD_NANDFLASH:
		printk("MTD_NANDFLASH:");
		break;
	case MTD_PEROM:
		printk("MTD_PEROM:");
		break;
	case MTD_OTHER:
		printk("MTD_OTHER:");
		break;
	case MTD_UNKNOWN:
		printk("MTD_UNKNOWN:");
		break;
	default:
		printk("Type %d:", mtd->type);
		break;
	}

	flags = mtd->flags;	/* don't toast the original */
	if ((flags & MTD_CLEAR_BITS)) {
		printk(" MTD_CLEAR_BITS");
		flags &= ~MTD_CLEAR_BITS;
	}
	if ((flags & MTD_SET_BITS)) {
		printk(" MTD_SET_BITS");
		flags &= ~MTD_SET_BITS;
	}
	if ((flags & MTD_ERASEABLE)) {
		printk(" MTD_ERASEABLE");
		flags &= ~MTD_ERASEABLE;
	}
	if ((flags & MTD_WRITEB_WRITEABLE)) {
		printk(" MTD_WRITEB_WRITEABLE");
		flags &= ~MTD_WRITEB_WRITEABLE;
	}
	if ((flags & MTD_VOLATILE)) {
		printk(" MTD_VOLATILE");
		flags &= ~MTD_VOLATILE;
	}
	if ((flags & MTD_XIP)) {
		printk(" MTD_XIP");
		flags &= ~MTD_XIP;
	}
	if ((flags & MTD_OOB)) {
		printk(" MTD_OOB");
		flags &= ~MTD_OOB;
	}
	if ((flags & MTD_ECC)) {
		printk(" MTD_ECC");
		flags &= ~MTD_ECC;
	}
	if (flags)
		printk(" ?%lx?", flags);
	printk("\n");

	printk("size %08lx erasesize %08lx bank_size %08lx\n", mtd->size, mtd->erasesize, mtd->bank_size);

	if (mtd->flags & MTD_OOB)
		printk("OOB block %08lx size %08lx\n", mtd->oobblock, mtd->oobsize);
	if (mtd->flags & MTD_ECC)
		printk("EEC type %08lx size %08lx\n", mtd->ecctype, mtd->eccsize);

	printk("Module %p\n", mtd->module);
	printk("Private %p (%p)\n", mtd->priv, mtd);

	printk("Funcs: read %p write %p erase %p\n", mtd->read, mtd->write, mtd->erase);
	printk("       sync %p point %p unpoint %p\n", mtd->sync, mtd->point, mtd->unpoint);
	if (mtd->flags & MTD_OOB)
		printk("       OOB read %p write %p\n", mtd->read_oob, mtd->write_oob);
	if (mtd->flags & MTD_ECC)
		printk("       ECC read %p write %p\n", mtd->read_ecc, mtd->write_ecc);
	printk("       PM suspend %p resume %p\n", mtd->suspend, mtd->resume);
}
#endif

/*
 * Local Variables:
 * compile-command: "gcc -DMODULE -D__KERNEL__ -O3 -Wall -c mtdnwflash.c"
 * End:
 */
