#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/sysctl.h>

static struct pci_dev *PMU = NULL;
static struct pci_dev *NB = NULL;

/* The status led GPIOs for v2 copper */
#define PMU_LED_REG		0xc3
static int netwinder_red_led = 1;
static int netwinder_green_led = 2;

static int netwinder_watchdog = 0;

/*
 * Given an input string in user space - get the integer value.
 * The string does not need to be null terminated.
 * Returns 0 if ok, error otherwise.
 * Only accepts positive decimal values.
 */
static int get_user_int(void *buffer, size_t size, void **end, int *value)
{
	char *s = buffer, *e, c;
	int i;

	/* Skip leading whitespace */
	for (i = 0; i < size; ++i, ++s) {
		if (get_user(c, s))
			return -EFAULT;
		if (c != ' ' && c != '\t')
			break;
	}

	/* Collect digits */
	*value = 0;
	for (e = s; i < size; ++i, ++e) {
		if (get_user(c, e))
			return -EFAULT;
		if (c < '0' || c > '9')
			break;
		*value = *value * 10 + c - '0';
	}

	/* No digits */
	if (e == s)
		return -EINVAL;

	if (end)
		*end = e;

	return 0;
}

static int nw_proc_cpu_handler(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp)
{
	int speed, rc = 0;
	char buf[8];

	if (!*lenp || (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}

	if (write) {
		rc = get_user_int(buffer, *lenp, NULL, &speed);
		if (rc)
			return rc;

		if (speed <= 25)
			speed = 0x15;
		else if (speed <= 38)
			speed = 0x17;
		else if (speed <= 50)
			speed = 0x19;
		else if (speed <= 63)
			speed = 0x1b;
		else if (speed <= 75)
			speed = 0x1d;
		else if (speed <= 88)
			speed = 0x1f;
		else /* if (speed <= 100) */
			speed = 0x01;

		pci_write_config_byte(NB, 0xa8, (u8) speed);
	} else {
		pci_read_config_byte(NB, 0xa8, (u8 *) &speed);
		switch (speed & 0x1e) {
		case 0x14:
			speed = 25;
			break;
		case 0x16:
			speed = 38;
			break;
		case 0x18:
			speed = 50;
			break;
		case 0x1a:
			speed = 63;
			break;
		case 0x1c:
			speed = 75;
			break;
		case 0x1e:
			speed = 88;
			break;
		default:
			speed = 100;
		}
		*lenp = sprintf(buf, "%d\n", speed) + 1;
		rc = copy_to_user(buffer, buf, *lenp);
		if (rc)
			return -EFAULT;
	}
	filp->f_pos += *lenp;
	return rc;
}

static int nw_proc_fanspeed_handler(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp)
{
	int speed, rc = 0, rpm = 0;
	char buf[16];

	if (!*lenp || (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}

	if (write) {
		rc = get_user_int(buffer, *lenp, NULL, &speed);
		if (rc)
			return rc;

		if (speed && speed <= 25)
			speed = 0x0d;
		else if (speed && speed <= 50)
			speed = 0x09;
		else if (speed && speed <= 75)
			speed = 0x05;
		else if (speed)
			speed = 0x01;

		pci_write_config_byte(PMU, 0xcb, (u8) speed);
	} else {
		pci_read_config_byte(PMU, 0xcb, (u8 *) &speed);
		switch (speed & 0x0d) {
		case 0x0d:
			speed = 25;
			break;
		case 0x09:
			speed = 50;
			break;
		case 0x05:
			speed = 75;
			break;
		case 0x01:
			speed = 100;
			break;
		default:
			speed = 0;
			break;
		}
		pci_read_config_byte(PMU, 0xca, (u8 *) &rpm);
		if (rpm > 1)
			*lenp = sprintf(buf, "%d %d\n", speed, 240000 / rpm) + 1;
		else
			*lenp = sprintf(buf, "%d %d\n", speed, 0) + 1;
		rc = copy_to_user(buffer, buf, *lenp);
		if (rc)
			return -EFAULT;
	}
	filp->f_pos += *lenp;
	return rc;
}

static int nw_proc_fan_handler(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp)
{
	char buf[2];
	u8 reg;
	int rc;

	if (!*lenp || (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}

	if (write) {
		char wason;
		int new;

		if ((rc = get_user_int(buffer, *lenp, NULL, &new)))
			return rc;
		if (new < 0 || new > 4)
			return -EINVAL;
		pci_read_config_byte(PMU, 0xcb, &reg);
		wason = reg & 1;
		reg &= ~0xd;
		switch (new) {
		case 0:
			break;
		case 1:
			/* The fan cannot start at the lowest setting */
			if (!wason) {
				pci_write_config_byte(PMU, 0xcb, reg | 1);
				current->state = TASK_INTERRUPTIBLE;
				schedule_timeout(2);
				current->state = TASK_RUNNING;
			}
			/* fall thru */
		default:
			reg |= ((4 - new) << 2) | 1;
			break;
		}
		pci_write_config_byte(PMU, 0xcb, reg);
	} else {
		pci_read_config_byte(PMU, 0xcb, &reg);
		if (reg & 1)
			buf[0] = 4 - ((reg >> 2) & 3) + '0';
		else
			buf[0] = '0';
		buf[1] = '\n';
		if (copy_to_user(buffer, buf, 2))
			return -EFAULT;
		*lenp = 2;
	}

	filp->f_pos += *lenp;
	return 0;
}

static int nw_proc_led_handler(ctl_table * table, int write, struct file *filp,
	void *buffer, size_t *lenp)
{
	char buf[2];
	u8 reg;

	if (!table->data || !table->maxlen || !*lenp
		|| (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}

	if (write) {
		int rc, new;

		if ((rc = get_user_int(buffer, *lenp, NULL, &new)))
			return rc;
		if (new < 0 && new > 1)
			return -EINVAL;
		pci_read_config_byte(PMU, PMU_LED_REG, &reg);
		if (new == 0)
			reg &= ~*(int *) table->data;
		else
			reg |= *(int *) table->data;
		pci_write_config_byte(PMU, PMU_LED_REG, reg);
	} else {
		pci_read_config_byte(PMU, PMU_LED_REG, &reg);
		buf[0] = (reg & *(int *) table->data) ? '1' : '0';
		buf[1] = '\n';
		if (copy_to_user(buffer, buf, 2))
			return -EFAULT;
		*lenp = 2;
	}

	filp->f_pos += *lenp;
	return 0;
}

static struct timer_list nw_timer;

/*  The watchdog is on GPIO1 */
static void nw_watchdog_timer(unsigned long data)
{
	u8 reg;

	/*
	 * This pulse must be at least 50ns long.  Since a PCI write takes
	 * 3 clocks (100ns), we do not need any delays. In reality, the
	 * PCI analyzer shows > 1us between accesses.
	 */
	pci_read_config_byte(PMU, 0x9c, &reg);
	pci_write_config_byte(PMU, 0x9c, reg | 2);
	pci_write_config_byte(PMU, 0x9c, reg & ~2);

	/*
	 * And do it again in a second and a half.  The watchdog barks at
	 * 1.6 seconds.
	 */
	nw_timer.expires = jiffies + HZ;
	add_timer(&nw_timer);
}

static void nw_stop_watchdog()
{
	u8 reg;

	pci_read_config_byte(PMU, 0x9a, &reg);
	pci_write_config_byte(PMU, 0x9a, reg & ~2);
	netwinder_watchdog = 0;
	del_timer(&nw_timer);
}

static int nw_proc_watchdog(ctl_table * table, int write, struct file *filp, void *buffer, size_t * lenp)
{
	char buf[2];
	u8 reg;

	if (!table->data || !table->maxlen || !*lenp
		|| (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}

	if (write) {
		int rc, new;

		if ((rc = get_user_int(buffer, *lenp, NULL, &new)))
			return rc;
		if (new < 0 && new > 1)
			return -EINVAL;
		if (new == netwinder_watchdog)
			return 0;
		netwinder_watchdog = new;

		if (new) {
			init_timer(&nw_timer);
			nw_timer.function = &nw_watchdog_timer;
			pci_read_config_byte(PMU, 0x9a, &reg);
			reg |= 2;
			pci_write_config_byte(PMU, 0x9a, reg);

			/* call it once to start things off */
			nw_watchdog_timer(0);
		} else
			nw_stop_watchdog();
	} else {
		buf[0] = netwinder_watchdog ? '1' : '0';
		buf[1] = '\n';
		if (copy_to_user(buffer, buf, 2))
			return -EFAULT;
		*lenp = 2;
	}

	filp->f_pos += *lenp;
	return 0;
}

/*  actual pins */
#define SPROM_CS	0x10
#define SPROM_CLK	0x20
#define SPROM_DI	0x40
#define SPROM_DO	0x80

/*  opcodes */
#define SPROM_READ		2

#define SPROM_SIZE		256

#define sprom_out(x)	\
	do { \
		pci_write_config_byte(PMU, 0x9e, reg | (x)); \
		pci_write_config_byte(PMU, 0x9e, reg | (x) | SPROM_CLK); \
		pci_write_config_byte(PMU, 0x9e, reg | (x)); \
	} while(0)

#define sprom_in(v) \
	do { \
		pci_write_config_byte(PMU, 0x9e, reg | SPROM_CLK); \
		pci_write_config_byte(PMU, 0x9e, reg); \
		pci_read_config_byte (PMU, 0x99, &v); \
	} while(0)

static void nw_sprom_cmd(int cmd, int addr, u8 reg)
{
	int i;

	reg |= SPROM_CS;

	/* start bit */
	sprom_out(SPROM_DI);

	/* opcode */
	sprom_out((cmd & 2) ? SPROM_DI : 0);
	sprom_out((cmd & 1) ? SPROM_DI : 0);

	/* address (8 bits - only lower 7 used) */
	for (i = 0x80; i > 0; i >>= 1) {
		sprom_out((addr & i) ? SPROM_DI : 0);
	}

	if (cmd == 0) {
		reg &= ~SPROM_CS;
		sprom_out(0);
	}
}

static int nw_sprom_read(int addr, char *buffer, int len)
{
	u8 reg, v;
	int i;
	u16 data;

	addr >>= 1;		/* convert from bytes to words */

	pci_read_config_byte(PMU, 0x9e, &reg);
	reg &= 0xf;
	reg |= SPROM_CS;

	nw_sprom_cmd(SPROM_READ, addr, reg);

	for (; len > 1; len -= 2, buffer += 2) {
		for (i = data = 0; i < 16; ++i) {
			sprom_in(v);
			data = (data << 1) | ((v & SPROM_DO) ? 1 : 0);
		}
		if (copy_to_user(buffer, &data, 2))
			return -EFAULT;
	}
	if (len) {
		for (i = data = 0; i < 16; ++i) {
			sprom_in(v);
			data = (data << 1) | ((v & SPROM_DO) ? 1 : 0);
		}
		if (copy_to_user(buffer, &data, 1))
			return -EFAULT;
	}

	/* done */
	reg &= ~SPROM_CS;
	sprom_out(0);

	return 0;
}

static int nw_proc_sprom(ctl_table * table, int write, struct file *filp,
	void *buffer, size_t * lenp)
{
	int rc;

	if (!*lenp || filp->f_pos >= SPROM_SIZE) {
		*lenp = 0;
		return 0;
	}
	if ((filp->f_pos & 1)) {
		printk("MISALIGNED len %x offset %lx\n", *lenp,
			(long) filp->f_pos);
		return -EINVAL;
	}

	if ((*lenp + filp->f_pos) > SPROM_SIZE) {
		*lenp = SPROM_SIZE - filp->f_pos;
	}

	if (write) {
		rc = -ENOSYS;
	} else {
		rc = nw_sprom_read(filp->f_pos, buffer, *lenp);
	}

	filp->f_pos += *lenp;
	return rc;
}

#define STD_CPUID 0x00000000
#define TM_CPUID 0x80860000

static void do_cpuid (u32 cpuid, u32 *eax, u32 *ebx, u32 *ecx, u32 *edx)
{
	asm("cpuid"
		: "=a" (*eax),
		  "=b" (*ebx),
		  "=c" (*ecx),
		  "=d" (*edx)
		: "a" (cpuid));
}

static int nw_proc_cms_handler(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp)
{
	char buf[80];
	u32 cms[2];
	u32 dummy1, dummy2;
	int rc;

	if (!*lenp || (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}

	do_cpuid(TM_CPUID+2, &dummy1, &cms[0], &cms[1], &dummy2);

	*lenp = sprintf(buf, "%d.%d.%d-%d-%d\n",
		(cms[0] >> 24) & 0xff, (cms[0] >> 16) & 0xff,
		(cms[0] >> 8) & 0xff, (cms[0] & 0xff), cms[1]);

	rc = copy_to_user(buffer, buf, *lenp);
	if (rc)
		return -EFAULT;

	filp->f_pos += *lenp;
	return rc;
}

static struct ctl_table_header *netwinder_table_header;

enum {
	DEV_NETWINDER_CPUSPEED = 1,
	DEV_NETWINDER_FANCTRL,
	DEV_NETWINDER_FANSPEED,
	DEV_NETWINDER_GREEN_LED,
	DEV_NETWINDER_RED_LED,
	DEV_NETWINDER_SPROM,
	DEV_NETWINDER_WATCHDOG,
	DEV_NETWINDER_CMS
};

enum {
	DEV_NETWINDER = 9	/* SAM HACK */
};

static ctl_table netwinder_table[] = {
	{ctl_name: DEV_NETWINDER_CPUSPEED, procname: "cpu_speed", mode:0644,
		proc_handler: nw_proc_cpu_handler},
	{ctl_name: DEV_NETWINDER_FANCTRL, procname: "fan_setting", mode: 0644,
		proc_handler: nw_proc_fan_handler},
	{ctl_name: DEV_NETWINDER_FANSPEED, procname: "fan_speed", mode: 0644,
		proc_handler: nw_proc_fanspeed_handler},
	{ctl_name: DEV_NETWINDER_RED_LED, procname: "red_led",
	 	data: &netwinder_red_led, maxlen: sizeof(int), mode: 0644,
	 	proc_handler: nw_proc_led_handler},
	{ctl_name: DEV_NETWINDER_GREEN_LED, procname: "green_led",
	 	data: &netwinder_green_led, maxlen: sizeof(int), mode: 0644,
	 	proc_handler: nw_proc_led_handler},
	{ctl_name: DEV_NETWINDER_WATCHDOG, procname: "watchdog",
	 	data: &netwinder_watchdog, maxlen: sizeof(int), mode: 0644,
	 	proc_handler: nw_proc_watchdog},
	{ctl_name: DEV_NETWINDER_SPROM, procname: "sprom", mode: 0644,
	 	proc_handler: nw_proc_sprom},
	{ctl_name: DEV_NETWINDER_CMS, procname: "cms", mode: 0444,
	 	proc_handler: nw_proc_cms_handler},
	{0}
};

static ctl_table netwinder_dir_table[] = {
	{ctl_name: DEV_NETWINDER, procname: "netwinder", mode: 0555,
	 child: netwinder_table},
	{0}
};

/* We assume that the leds have been initialized */
int init_module(void)
{
	u8 reg;

	/* find the PMU */
	if (!(PMU = pci_find_slot(0, PCI_DEVFN(0x11, 0)))) {
		printk("You do not have a PMU\n");
		return -ENODEV;
	}

	/* find the Northbridge (NB) */
	if (!(NB = pci_find_slot(0, PCI_DEVFN(0, 0)))) {
		printk("You do not have a northbridge\n");
		return -ENODEV;
	}

	/* setup sprom */
	pci_read_config_byte(PMU, 0x9f, &reg);
	reg = (reg & 0xf) | 0x70;
	pci_write_config_byte(PMU, 0x9f, reg);
	pci_read_config_byte(PMU, 0x9e, &reg);
	reg = reg & 0xf;
	pci_write_config_byte(PMU, 0x9e, reg);

	netwinder_table_header = register_sysctl_table(netwinder_dir_table, 1);

	return 0;
}

void cleanup_module(void)
{
	/* Make sure the watchdog is off. */
	if (netwinder_watchdog)
		nw_stop_watchdog();

	unregister_sysctl_table(netwinder_table_header);
}

/*
 * Local Variables:
 * compile-command: "gcc -O3 -Wall -D__KERNEL__ -DMODULE -c sysnetwinder.c"
 * End:
 */
